/******************************************************************************
*******************************************************************************


	SAT (separating axis theorem) test DS
	relminator (Richard Eric M. Lope BSN RN)
	http://rel.betterwebber.com


	
*******************************************************************************
******************************************************************************/

#include <vector>

#include <nds.h>
#include <stdio.h>
#include <math.h>

#include "Cvector2df32.h"

#include "Cglsprite.h"
#include "Cgl2d.h"


using namespace std;

class Cprojectionf32
{
	public:
	s32		min;
	s32 	max;
};


class Cpolygonf32
{
	public:
	Cprojectionf32 project( Cvector2df32 _axis );
	void create( int _numverts, int radius );
	void draw( Cgl2d &gl2d, int color );
	void move( int _x, int _y);
	int						num_verts;
	s32						x;
	s32						y;
	vector<Cvector2df32> 	v_vertex;
	vector<Cvector2df32> 	v_coords;
	
};

Cprojectionf32 Cpolygonf32::project( Cvector2df32 _axis )
{

	Cprojectionf32 proj;
	s32 min, max;
	s32 p;
	
	// get initial projection on first vertex
	// and set min and max values to initial projection value
	p = vector2df32_dot(v_coords[0], _axis);
	min = p;
	max = p;
	
	// loop through all the verts of the poly 
	// and find out if new projections of vertex
	// is the new min or max as we need only the 
	// min and max points to test for 1d intesection
	for (int i = 1; i < num_verts; i++)
	{
		p = vector2df32_dot(v_coords[i], _axis);
		if (p < min) min = p;
		if (p > max) max = p;
	}

	// copy and return
	proj.min = min;
	proj.max = max;
	
	return (proj);
	
}

void Cpolygonf32::create( int _numverts, int radius )
{

	num_verts = _numverts;
	v_vertex.resize(num_verts);
	v_coords.resize(num_verts);

	// verts
	for (int i = 0; i < num_verts; i++)
	{
		s32 angle = (32768/num_verts) * i;
		v_vertex[i].x = cosLerp(angle) * radius;
		v_vertex[i].y = sinLerp(angle) * radius;
	}
	
}

void Cpolygonf32::draw( Cgl2d &gl2d, int color )
{

	for (int i = 0; i < num_verts; i++)
	{
		int j = (i + 1) % num_verts;
		int x1 = v_coords[i].x >> 12;
		int y1 = v_coords[i].y >> 12;
		int x2 = v_coords[j].x >> 12;
		int y2 = v_coords[j].y >> 12;
		
		gl2d.line(x1, y1, x2, y2, color);
		
		// normal
		
		x1 = v_coords[i].x;
		y1 = v_coords[i].y;
		x2 = v_coords[j].x;
		y2 = v_coords[j].y;
		
		Cvector2df32 axis = line_normalf32(x1, y1, x2, y2, false);
		vector2df32_normalize(axis);
		s32 mx,my;
		
		mx = (x2+x1)/2;
		my = (y1+y2)/2;
		
		x1 = mx >> 12;
		y1 = my >> 12;
		x2 = (mx + axis.x*20) >> 12;
		y2 = (my + axis.y*20) >> 12;
		
		gl2d.line(x1, y1, x2, y2, color);
		
	}
	
}

void Cpolygonf32::move( s32 _x, s32 _y )
{

	for (int i = 0; i < num_verts; i++)
	{
		v_coords[i].x = _x + v_vertex[i].x;
		v_coords[i].y = _y + v_vertex[i].y;
	}
}

bool poly_collide(Cpolygonf32 &p1, Cpolygonf32 &p2)
{
	
	Cprojectionf32 proj1, proj2;
	Cvector2df32 axis;
	s32 d1, d2;
	s32 x1, y1, x2,y2;
	// project all the verts of the poly to each axis (normal)
	// of the poly we are testing and find out if the projections
	// overlap (ie: length if proj1 and proj2 are intersecting).
	// if they are intersecting, there is an axis (line perpendicular 
	// to the axis tested or the "edge" of the poly where the normal connects)
	// that separates the two polygons so we do an early out from the function.
	
	
	// polygon1
	for (int i = 0; i < p1.num_verts; i++)
	{
		int j = (i + 1) % p1.num_verts;
		x1 = p1.v_coords[i].x;
		y1 = p1.v_coords[i].y;
		x2 = p1.v_coords[j].x;
		y2 = p1.v_coords[j].y;
		
		axis = line_normalf32(x1, y1, x2, y2, false);
		//vector2df32_normalize(axis);			
		
		proj1 = p1.project(axis);
		proj2 = p2.project(axis);
		d1 = (proj1.min - proj2.max);
	    d2 = (proj2.min - proj1.max);
	    if ((d1 > 0) || (d2 > 0))      // there's a separating axis so get out early
	        return  (false);
	    
	}	
	
	// polygon2
	for (int i = 0; i < p2.num_verts; i++)
	{
		int j = (i + 1) % p2.num_verts;
		x1 = p2.v_coords[i].x;
		y1 = p2.v_coords[i].y;
		x2 = p2.v_coords[j].x;
		y2 = p2.v_coords[j].y;
		
		axis = line_normalf32(x1, y1, x2, y2, false);
		//vector2df32_normalize(axis);
		
		proj1 = p1.project(axis);
		proj2 = p2.project(axis);
		d1 = (proj1.min - proj2.max);
	    d2 = (proj2.min - proj1.max);
	    if ((d1 > 0) || (d2 > 0))      // there's a separating axis so get out early
	        return  (false);
	    
	}	
	
	
	// no separating axis found so p1 and p2 are colliding
	return (true);
	
}

bool poly_collide_mtv(Cpolygonf32 &p1, Cpolygonf32 &p2, Cvector2df32 &mtv)
{
	
	Cprojectionf32 proj1, proj2;
	Cvector2df32 axis;
	s32 d1, d2, magnitude, overlap;
	s32 x1, y1, x2,y2;
	
	// project all the verts of the poly to each axis (normal)
	// of the poly we are testing and find out if the projections
	// overlap (ie: length if proj1 and proj2 are intersecting).
	// if they are intersecting, there is an axis (line perpendicular 
	// to the axis tested or the "edge" of the poly where the normal connects)
	// that separates the two polygons so we do an early out from the function.
	
	
	magnitude = 0xFFFFFFFF;
	// polygon1
	for (int i = 0; i < p1.num_verts; i++)
	{
		int j = (i + 1) % p1.num_verts;
		x1 = p1.v_coords[i].x;
		y1 = p1.v_coords[i].y;
		x2 = p1.v_coords[j].x;
		y2 = p1.v_coords[j].y;
		
		axis = line_normalf32(x1, y1, x2, y2, false);
		vector2df32_normalize(axis);
		
		proj1 = p1.project(axis);
		proj2 = p2.project(axis);
		d1 = (proj1.min - proj2.max);
	    d2 = (proj2.min - proj1.max);
	    if ((d1 > 0) || (d2 > 0))      // there's a separating axis so get out early
		{
			if(d1 < d2)
	    		overlap = d1;
	    	else
	    		overlap = d2;
	    	
	    
	    	if (overlap < magnitude)
			{
	    		magnitude = overlap;
				Cvector2df32 vd;
	    		vd.x = p1.x - p2.x;
				vd.y = p1.y - p2.y;
	    		if (vector2df32_dot(vd, axis) < 0)
				{
	    			mtv.x = -axis.x;
					mtv.y = -axis.y;
	    		}
				else
	    		{
	    			mtv.x = axis.x;
					mtv.y = axis.y;
	    		}	
	    	} 
		 
			return  (false);
	    }
	}	
	
	// polygon2
	for (int i = 0; i < p2.num_verts; i++)
	{
		int j = (i + 1) % p2.num_verts;
		x1 = p2.v_coords[i].x;
		y1 = p2.v_coords[i].y;
		x2 = p2.v_coords[j].x;
		y2 = p2.v_coords[j].y;
		
		axis = line_normalf32(x1, y1, x2, y2, false);
		vector2df32_normalize(axis);
		
		proj1 = p1.project(axis);
		proj2 = p2.project(axis);
		d1 = (proj1.min - proj2.max);
	    d2 = (proj2.min - proj1.max);
	    if ((d1 > 0) || (d2 > 0))      // there's a separating axis so get out early
		{
			if(d1 < d2)
	    		overlap = d1;
	    	else
	    		overlap = d2;
	    	
	    
	    	if (overlap < magnitude)
			{
	    		magnitude = overlap;
				Cvector2df32 vd;
	    		vd.x = p1.x - p2.x;
				vd.y = p1.y - p2.y;
	    		if (vector2df32_dot(vd, axis) < 0)
				{
	    			mtv.x = -axis.x;
					mtv.y = -axis.y;
	    		}
				else
	    		{
	    			mtv.x = axis.x;
					mtv.y = axis.y;
	    		}	
	    	} 
		 
			return  (false);
	    }
	}	
	
	
	// no separating axis found so p1 and p2 are colliding
	return (true);
	
}


void init_gl_screen();

Cgl2d gl2d;

Cpolygonf32 p1,p2;


int main()
{
	
	p1.create(4, 20);
	p1.move( 50 << 12, 50 << 12);

	p2.create(5, 30);
	p2.move( 120 << 12, 100 << 12);

	videoSetMode(MODE_5_3D);
	
	
	consoleDemoInit();
	
	iprintf("\x1b[1;1HSAT Test DS");
	iprintf("\x1b[2;1HRelminatorS");
	iprintf("\x1b[3;1Hhttp://rel.betterwebber.com");

	init_gl_screen();
	
	int key;

	s32 x = 50 << 12;
	s32 y = 50 << 12;
	
	static const s32 speed = 1 << 12;
	s32 dx = speed << 12; 
	s32 dy = speed << 12; 
	
	Cvector2df32 mtv;
	
	while(1)
	{
	
		scanKeys();
		key = keysHeld();
		
		dx = 0;
		dy = 0;
		
		if (key & KEY_RIGHT)
		{
			dx += speed;
		}
		
		if (key & KEY_LEFT)
		{
			dx -= speed;
		}
		
		if (key & KEY_UP)
		{
			dy -= speed;
		}
		
		if (key & KEY_DOWN)
		{
			dy += speed;
		}
		
		
		bool c = poly_collide_mtv(p1,p2,mtv);
		
		if (c)
		{
			x -= (dx-mtv.x);
			y -= (dy-mtv.y);
		}
		else
		{
			x += (dx);
			y += (dy);
		}
		
		
		p1.move(x, y);
		
		
		
		// start 2d mode
		gl2d.start();
			
			// needs to be called before texturing
			// or sprite gets blended by last glcolor call
			glColor(ARGB16(1,31,31,31)); 
			
			p1.draw(gl2d, RGB15(31,31,31));
			p2.draw(gl2d, RGB15(0,31,31));
			
			//gl2d.box_fill(10,10, 30, 50, RGB15(31,0,0));
			
			
		gl2d.end();
		
		
		
		
	
		glFlush(0);

		swiWaitForVBlank();
		
		iprintf("\x1b[5;1Hcollision = %i ", c);
		
	}

	
	return 0;
	
}



void init_gl_screen()
{

	// initialize gl
	glInit();
	
	//enable textures
	glEnable(GL_TEXTURE_2D);
	
	// enable antialiasing
	glEnable(GL_ANTIALIAS);
		
	// setup the rear plane
	glClearColor(0,0,0,31); // BG must be opaque for AA to work
	glClearPolyID(63); // BG must have a unique polygon ID for AA to work
	
	glClearDepth(GL_MAX_DEPTH);

	//this should work the same as the normal gl call
	glViewport(0,0,255,191);
	
	
	//any floating point gl call is being converted to fixed prior to being implemented
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70, 256.0 / 192.0, 1, 200);
	
	gluLookAt(	0.0, 0.0, 1.0,		//camera possition 
				0.0, 0.0, 0.0,		//look at
				0.0, 1.0, 0.0);		//up

	glMaterialf(GL_AMBIENT, RGB15(31,31,31));
	glMaterialf(GL_DIFFUSE, RGB15(31,31,31));
	glMaterialf(GL_SPECULAR, BIT(15) | RGB15(31,31,31));
	glMaterialf(GL_EMISSION, RGB15(31,31,31));

	//ds uses a table for shinyness..this generates a half-ass one
	glMaterialShinyness();

	//not a real gl function and will likely change
	glPolyFmt(POLY_ALPHA(31) | POLY_CULL_BACK);
}

